SystemOrganization addCategory: #'RPackage-Core'!

!ClassDescription methodsFor: '*rpackage-core' stamp: 'sd 10/29/2009 20:31'!
compileSilently: code 
	"Compile the code and classify the resulting method in the given category, leaving no trail in the system log, nor in any change set, nor in the 'recent submissions' list. This should only be used when you know for sure that the compilation will succeed."

	^ self compileSilently: code classified: 'not defined category' notifying: nil.! !

!ClassDescription methodsFor: '*rpackage-core' stamp: 'tg 6/13/2010 17:32'!
extendingPackages
	"the extending packages of a class are the packages that extend it."
	^ RPackage organizer extendingPackagesOf: self
	! !

!ClassDescription methodsFor: '*rpackage-core' stamp: 'StephaneDucasse 4/20/2010 22:10'!
isDefinedInPackage: aPackage
	"returns true if aPackage contains the definitino of this class"
	
	^ aPackage includesClass: self.
	! !

!ClassDescription methodsFor: '*rpackage-core' stamp: 'StephaneDucasse 5/15/2010 19:54'!
isExtended

	^ self extendingPackages isEmpty! !

!ClassDescription methodsFor: '*rpackage-core' stamp: 'tg 3/16/2010 17:53'!
isExtendedInPackage: aPackage
	"returns true if aPackage defines an extension to this class"
	^ aPackage extendsClass: self.
	! !

!ClassDescription methodsFor: '*rpackage-core' stamp: 'tg 6/13/2010 17:19'!
package
	^ RPackage organizer packageOf: self.	! !

!ClassDescription methodsFor: '*rpackage-core' stamp: 'tg 3/16/2010 17:36'!
packageFromOrganizer: anOrganizer
	"returns the package that defines this class"
	^ anOrganizer packageOf: self.
	! !

!ClassDescription methodsFor: '*rpackage-core' stamp: 'sd 5/19/2010 16:31'!
packages
	"the extending packages of a class are the packages that extend it."
	
	^  self extendingPackages asSet copy add: self package; yourself! !

!CompiledMethod methodsFor: '*rpackage-core' stamp: 'tg 3/13/2010 15:13'!
isDefinedInPackage: anRPackage
	^ anRPackage includesDefinedSelector: self selector ofClass: self methodClass! !

!CompiledMethod methodsFor: '*rpackage-core' stamp: 'tg 3/13/2010 13:50'!
isExtensionInPackage: anRPackage
	^ anRPackage includesExtensionSelector: self selector ofClass: self methodClass! !

!CompiledMethod methodsFor: '*rpackage-core' stamp: 'tg 3/13/2010 16:37'!
packageFromOrganizer: anRPackageOrganizer
	^ anRPackageOrganizer packageDefiningOrExtendingMethod: self! !

Object subclass: #RPackage
	instanceVariableNames: 'classDefinedSelectors metaclassDefinedSelectors classExtensionSelectors metaclassExtensionSelectors classes name'
	classVariableNames: 'PackageGlobalOrganizer'
	poolDictionaries: ''
	category: 'RPackage-Core'!

!RPackage commentStamp: 'sd 4/25/2010 17:49' prior: 0!
A RPackage2 is a new implementation of package. Note that I does not touch classes but I'm just a layer on top of classes therefore removing a method from a package does not change the underlying class. 


Instance Variables
	classDefinedSelectors:		Dictionary
	classExtensionSelectors:		Dictionary
	metaclassDefinedSelectors:		Dictionary
	metaclassExtensionSelectors:		Dictionary
	packageName:		String
	classes: OrderedCollection of Symbols

The reason we use four dictionaries is that this way we only store #Point in both as the class and the metaclass are involved. We do not have to with Point_class problem when comparing defined and extended classes. The idea is taken from the RBSelectorEnvironment. 					
		
Class Invariant one:
	Classes should not contain metaclass name, but only class names
	
	

Implementation notes 
====================
This class went over 3 internal representations implementation. 
	- first: 	a list of class + two dictionaries: class * methods
	This was not good since we add to go all the time over the list of classes.
	- second: 4 dictionaries class * selectors 
	This was not good since we want to have defined classes without methods. 
	- third: 4 dictionaries + definedClasses

Originally I wanted to avoid to have a defined class list and I wanted to avoid to have to declare the class as defined. But this is not really good since
	- we will want to know if a class definition (without method for example is defined in a given package)
	- second this is easier to hook the packageOf: behavior (we register the package class when the class is 	added to the package).

defined classes information is redundant with the dictionary keys of defined methods but we would have to check and register the class to the packageOrganizer the first time a method is defined then also check on remove to unregister the class from the packageOrganizer.

Adding a method does not define the class as a defined package class. This has to be done explictly. The reason for this choice is that a class should register to the packageOrganizer and that I do not want to do it each time a method is added and I do not want to test it each time a method is added. Now this is done only when the class is declared as defined. 
We could also give the complete freedom to the client to register the class but I thought it was a good compromise. 
	
Todo
====
	Next: 
	- should use resources for the completeSetupTest
	- register to systemNotifier
	- build up a synchronizer that import PackageInfo.
	We should port all the tests.
	We should test packages and extendingPackages defined on classDescription
	We should add MC support.
	- add events.

	

	
!

!RPackage class methodsFor: 'initialize-release' stamp: 'tg 6/13/2010 17:41'!
initialize
	"self initialize"
	self organizer: RPackageOrganizer new! !

!RPackage class methodsFor: 'instance creation' stamp: 'tg 6/13/2010 17:22'!
named: aString

	| newPackage |
	newPackage := self new name: aString ;yourself.
	self organizer registerPackage: newPackage.
	^ newPackage! !

!RPackage class methodsFor: 'protected' stamp: 'tg 6/13/2010 16:59'!
organizer
	^ PackageGlobalOrganizer! !

!RPackage class methodsFor: 'protected' stamp: 'tg 6/13/2010 17:00'!
organizer: aPackageOrganizer
	PackageGlobalOrganizer := aPackageOrganizer! !

!RPackage class methodsFor: 'protected' stamp: 'tg 6/13/2010 17:01'!
withOrganizer: aNewOrganizer do: aBlock
	| old |
	[
		old := self organizer.
		self organizer: aNewOrganizer.
		aBlock value.
	] ensure: [ self organizer: old ]! !

!RPackage methodsFor: 'class addition' stamp: 'sd 4/22/2010 07:58'!
addClassDefinition: aClass
	"Add only class definition, aClass, as defined classes of the receiver. The term classdefinition denotes the fact that we do not treat its methods. aClass methods are not added to the package. aClass can be either a class or a metaclass"
	
	self addClassDefinitionName: aClass theNonMetaClass name.
	! !

!RPackage methodsFor: 'class addition' stamp: 'sd 4/24/2010 22:47'!
addClassDefinitionName: aClassName
	"aClassName should only be a class (and not a metaclass name)"
	
	| aClassNameSymbol |
	('* class' match: aClassName) ifTrue: [^ self error: 'no metaclass name'].
	aClassNameSymbol := aClassName asSymbol.
	(classes includes: aClassNameSymbol)
		ifFalse: 
			[classes add: aClassNameSymbol.
			self extendedMethodsShouldBecomeDefinedWhenAddingClassName: aClassNameSymbol.
			self registerClassName: aClassNameSymbol]
	! !

!RPackage methodsFor: 'compiled method handling' stamp: 'tg 6/13/2010 17:44'!
addMethod: aCompiledMethod 
	"Add the method to the receiver as a defined method if the class is  defined in it, else as an extension."

	| mthClass |
	mthClass := aCompiledMethod methodClass. 
	(self includesClass: mthClass )
		ifFalse: [mthClass isMeta
					ifTrue: [(metaclassExtensionSelectors 
								at: mthClass theNonMetaClass name  
								ifAbsentPut: [Set new]) add: aCompiledMethod selector.]
					ifFalse: [(classExtensionSelectors 
								at: mthClass name 
								ifAbsentPut: [Set new]) add: aCompiledMethod selector].
				"we added a method extension so the receiver is an extending package of the class"
				RPackage organizer registerExtendingPackage: self forClass: mthClass]
		ifTrue: [mthClass isMeta
					ifTrue: [(metaclassDefinedSelectors 
								at: mthClass theNonMetaClass name  
								ifAbsentPut: [Set new]) add: aCompiledMethod selector]
					ifFalse: [(classDefinedSelectors 
								at: mthClass name 
								ifAbsentPut: [Set new]) add: aCompiledMethod selector]].	
	^ aCompiledMethod.! !

!RPackage methodsFor: 'selector handling' stamp: 'StephaneDucasse 4/20/2010 22:56'!
addSelector: aSelector ofClassName: aClassName
	"Add the method to the receiver. If the class is not locally defined in that package then the method is defined as a method extension: ie extending another package. Note that this method does not add the method to the class, it just records in the package that the method is packaged."
	
	( self includesClassNamed: aClassName)
		ifFalse: [(classExtensionSelectors at: aClassName asSymbol ifAbsentPut: [Set new]) add: aSelector]
		ifTrue: [(classDefinedSelectors at: aClassName asSymbol ifAbsentPut: [Set new]) add: aSelector].
	^ aSelector.! !

!RPackage methodsFor: 'selector handling' stamp: 'StephaneDucasse 4/20/2010 22:56'!
addSelector: aSelector ofMetaclassName: aClassName
	"Add the method to the receiver. If the class is not locally defined in that package then the method is defined as a method extension: ie extending another package. Note that this method does not add the method to the class, it just records in the package that the method is packaged. aClassName is the sole instance class name and not its metaclass one: i.e. adding Point class>>new is done as addSelector: #new ofMetaclassName: #Point"
	
	( self includesClassNamed: aClassName) 
		ifFalse: [(metaclassExtensionSelectors at: aClassName asSymbol ifAbsentPut: [Set new]) add: aSelector]
		ifTrue: [(metaclassDefinedSelectors at: aClassName asSymbol ifAbsentPut: [Set new]) add: aSelector].
	^ aSelector.! !

!RPackage methodsFor: 'slices' stamp: 'StephaneDucasse 3/11/2010 18:46'!
classDefinedSlicesDo: aBlock
	"This method iterates over the defined class and their associated selectors. a slice is a class * list of selectors. aBlock will be applied to all the extensions slices of the receiver. aBlok first argument is the class and the second argument a list of method selectors"
	
	classDefinedSelectors keysAndValuesDo: aBlock! !

!RPackage methodsFor: 'slices' stamp: 'StephaneDucasse 3/11/2010 18:46'!
classExtensionSlicesDo: aBlock
	"This method iterates over the class extensions and their associated selectors. A slice is a class * list of selectors. aBlock will be applied to all the extensions slices of the receiver. aBlok first argument is the class and the second argument a list of method selectors"
	
	classExtensionSelectors keysAndValuesDo: aBlock! !

!RPackage methodsFor: 'slices' stamp: 'StephaneDucasse 3/11/2010 18:40'!
classNamesAndExtensionSelectorsDo: aBlock
	"Iterate over the extended methods grouped by classes and apply the argument.
	The block will be passed a class name and each individual selectors.
	Only classes with methods are paired with their methods"
	
	classExtensionSelectors keysAndValuesDo: [ :classSymbol :methods |
			methods do: [ :sel |  aBlock value: classSymbol	value: sel]].
	
	metaclassExtensionSelectors keysAndValuesDo: [ :classSymbol :methods |
			methods do: [ :sel |  aBlock 
									value: (classSymbol, '_class') asSymbol
									value: sel]].	! !

!RPackage methodsFor: 'all accessing' stamp: 'sd 4/25/2010 16:03'!
classes
	"Return all the classes"
	
	^ self definedClasses, self extendedClasses 
	
	! !

!RPackage methodsFor: 'all accessing' stamp: 'StephaneDucasse 3/11/2010 15:17'!
definedClassNames 
	"Return the class names having methods defined in the receiver."
	
	^ classes! !

!RPackage methodsFor: 'all accessing' stamp: 'StephaneDucasse 5/15/2010 19:47'!
definedClasses 
	
	^ classes collect: [:each | Smalltalk globals at: each]
	! !

!RPackage methodsFor: 'iteration' stamp: 'StephaneDucasse 2/21/2010 17:25'!
definedClassesDo: aBlock

	classDefinedSelectors keysDo: aBlock.
	metaclassDefinedSelectors keysDo: aBlock.! !

!RPackage methodsFor: 'private' stamp: 'sd 4/25/2010 15:29'!
definedMethodsBecomeExtendedForClass: aClassName
	"the package may contain defined methods and their class is removed to the receiver. The status of such methods should now be extended"
	"Precondition: aClassName is in the defined list"
	
	| ext |
	(classDefinedSelectors keys includes: aClassName) ifFalse: [^ self].
	ext := classDefinedSelectors at: aClassName.
	classDefinedSelectors removeKey: aClassName.
	(classExtensionSelectors keys includes: aClassName) ifTrue: [self error: aClassName , ' should not be defined in extension'.]. "we should remove this check later"
	classExtensionSelectors at: aClassName put: ext.! !

!RPackage methodsFor: 'private' stamp: 'sd 4/25/2010 15:37'!
definedMethodsBecomeExtendedForMetaclass: aClassName
	"the package may contain defined methods and their class is removed to the receiver. The status of such methods should now be extended"
	"Precondition: aClassName is in the defined list"
	
	| ext |
	(metaclassDefinedSelectors keys includes: aClassName) ifFalse: [^ self].
	ext := metaclassDefinedSelectors at: aClassName.
	metaclassDefinedSelectors classDefinedSelectors removeKey: aClassName.
	(metaclassExtensionSelectors keys includes: aClassName) ifTrue: [self error: aClassName , ' should not be defined in extension'.]. "we should remove this check later"
	metaclassDefinedSelectors at: aClassName put: ext.! !

!RPackage methodsFor: 'accessing' stamp: 'tg 3/13/2010 14:17'!
definedMethodsForClass: aClass

	^ (self definedSelectorsForClass: aClass) collect: [ :each | aClass >> each ]! !

!RPackage methodsFor: 'private' stamp: 'sd 4/25/2010 15:41'!
definedMethodsShouldBecomeExtensionWhenRemovingClass: aClassName
	"the package may contain defined methods and their class is removed from the receiver. The status of such methods should now be extensions"
	
	self definedMethodsBecomeExtendedForClass: aClassName.
	self definedMethodsBecomeExtendedForMetaclass: aClassName.! !

!RPackage methodsFor: 'accessing' stamp: 'tg 3/13/2010 13:31'!
definedSelectorsForClass: aClass

	^ aClass isMeta
		ifTrue: [ metaclassDefinedSelectors at: aClass theNonMetaClass name ifAbsent: [ #() ] ]
		ifFalse: [ classDefinedSelectors at: aClass name ifAbsent: [ #() ] ]! !

!RPackage methodsFor: 'testing' stamp: 'StephaneDucasse 4/20/2010 22:22'!
definesClass: aClass
	"Returns true if the class, aClass, is one of the locally defined classes"
	"should be probably removed since this is the same as includesClass: or the inverse"
	^ self includesClass: aClass! !

!RPackage methodsFor: 'testing' stamp: 'sd 4/25/2010 16:06'!
definesOrExtendsClass: aClass
	"Returns true whether the class, aClass, is one of the locally defined classes of the receiver or 
	if the receiver extends such class (that is defined in another package)"
	
	| canonizedName |
	canonizedName := aClass theNonMetaClass name.
	^ (classes includes: canonizedName) 
		or: [self extendedClassNames includes:  canonizedName]
! !

!RPackage methodsFor: 'accessing' stamp: 'sd 4/25/2010 16:06'!
extendedClassNames 
	"Return the name of the classes which are extended by the receiver package. if a metaclass is extended, just get its sole instance class name."
	
	^ classExtensionSelectors keys union: metaclassExtensionSelectors keys! !

!RPackage methodsFor: 'accessing' stamp: 'sd 4/25/2010 16:03'!
extendedClasses 
	"Return classes and metaclasses that are extended in the receiver"
	
	^ (metaclassExtensionSelectors keys collect: [:each | (Smalltalk at: each) class])
			union:  (classExtensionSelectors keys collect: [:each | Smalltalk at: each])! !

!RPackage methodsFor: 'private' stamp: 'sd 4/25/2010 15:29'!
extendedMethodsBecomeDefinedForClass: aClassName
	"the package may contain extension methods and their class is added to the receiver. The status of such methods should now be defined"
	"Precondition: aClassName is in the defined list"
	
	| ext |
	(classExtensionSelectors keys includes: aClassName) ifFalse: [^ self].
	ext := classExtensionSelectors at: aClassName.
	classExtensionSelectors removeKey: aClassName.
	(classDefinedSelectors keys includes: aClassName) ifTrue: [self error: aClassName , ' should not be defined in extension'.]. "we should remove this check later"
	classDefinedSelectors at: aClassName put: ext.! !

!RPackage methodsFor: 'private' stamp: 'StephaneDucasse 5/15/2010 22:22'!
extendedMethodsBecomeDefinedForMetaclass: aClassName
	"the package may contain extension methods and their class is added to the receiver. The status of such methods should now be defined"
	"Precondition: aClassName is in the defined list"
	
	| ext |
	(metaclassExtensionSelectors keys includes: aClassName) ifFalse: [^ self].
	ext := metaclassExtensionSelectors at: aClassName.
	metaclassExtensionSelectors removeKey: aClassName.
	(metaclassDefinedSelectors keys includes: aClassName) 
			ifTrue: [self error: aClassName , ' should not be defined in extension'.]. "we should remove this check later"
	metaclassDefinedSelectors at: aClassName put: ext.! !

!RPackage methodsFor: 'private' stamp: 'sd 4/25/2010 15:29'!
extendedMethodsShouldBecomeDefinedWhenAddingClassName: aClassName
	"the package may contain extension methods and their class is added to the receiver. The status of such methods should now be defined"
	"Precondition: aClassName is in the defined list"
	
	self extendedMethodsBecomeDefinedForClass: aClassName.
	self extendedMethodsBecomeDefinedForMetaclass: aClassName.
	! !

!RPackage methodsFor: 'testing' stamp: 'sd 4/25/2010 16:05'!
extendsClass: aClass
	"Returns true if the receiver extends aClass (that is defined in another package)"
	
	| canonizedName |
	canonizedName := aClass theNonMetaClass name.
	^ self extendedClassNames includes:  canonizedName! !

!RPackage methodsFor: 'accessing' stamp: 'sd 4/25/2010 17:29'!
extensionMethods
	"extension methods are methods defined on classes that are not defined on the receiver"
	
	| mthds |
	mthds := OrderedCollection new.
	classExtensionSelectors keysAndValuesDo: [ :classSymbol :methods |
			methods do: [ :sel | mthds add: ((Smalltalk at: classSymbol) >>	sel)]].
	metaclassExtensionSelectors keysAndValuesDo: [ :classSymbol :methods |
			methods do: [ :sel | mthds add: ((Smalltalk at: classSymbol) >>	sel)]].	
	^ mthds				! !

!RPackage methodsFor: 'accessing' stamp: 'tg 3/13/2010 14:18'!
extensionMethodsForClass: aClass

	^ (self extensionSelectorsForClass: aClass) collect: [ :each | aClass >> each ]! !

!RPackage methodsFor: 'accessing' stamp: 'sd 4/25/2010 17:07'!
extensionSelectors
	"extension methods are methods defined on classes that are not defined on the receiver"
	
	| classSels |
	classSels := Set new.
	classExtensionSelectors values do: [:each | classSels addAll: each].
	metaclassExtensionSelectors values do: [:each | classSels addAll: each].
	^ classSels.
	! !

!RPackage methodsFor: 'accessing' stamp: 'StephaneDucasse 4/21/2010 14:05'!
extensionSelectorsForClass: aClass

	^ aClass isMeta
		ifTrue: [ metaclassExtensionSelectors at: aClass theNonMetaClass name ifAbsent: [ #() ] ]
		ifFalse: [ classExtensionSelectors at: aClass name ifAbsent: [ #() ] ]! !

!RPackage methodsFor: 'testing' stamp: 'StephaneDucasse 2/21/2010 17:56'!
includesClass: aClass	
	"Returns true if the receiver includes aClass in the classes that are defined within it: only class definition are considered - not class extensions"
	
	^ self includesClassNamed: aClass theNonMetaClass name! !

!RPackage methodsFor: 'testing' stamp: 'StephaneDucasse 4/20/2010 22:53'!
includesClassNamed: aSymbol
	"Returns true if the receiver includes class named aSymbol in the classes that are defined within it: only class definition are considered - not class extensions"
	
	| aRealSymbol |
	aRealSymbol := aSymbol asSymbol.
	^ classes includes: aSymbol
	! !

!RPackage methodsFor: 'testing' stamp: 'tg 3/13/2010 15:13'!
includesDefinedSelector: aSelector ofClass: aClass
	"Return true if the receiver includes the method of selector aSelector. Only checks methods defined in this package"
	
	| sels |
	sels := self definedSelectorsForClass: aClass.
	^ sels includes: aSelector.! !

!RPackage methodsFor: 'testing' stamp: 'StephaneDucasse 2/21/2010 18:53'!
includesDefinedSelector: aSelector ofClassName: aClassName
	"Return true if the receiver includes the method of selector aSelector. Only checks methods defined (not extended by other packages or package extensions)"
	
	| sels |
	sels := classDefinedSelectors at: aClassName ifAbsent: [^ false].
	^ sels includes: aSelector.! !

!RPackage methodsFor: 'testing' stamp: 'StephaneDucasse 2/21/2010 18:53'!
includesDefinedSelector: aSelector ofMetaclassName: aClassName
	"Return true if the receiver includes the method of selector aSelector. Only checks methods defined (not extended by other packages or package extensions)"
	
	| sels |
	sels := metaclassDefinedSelectors at: aClassName ifAbsent: [^ false].
	^ sels includes: aSelector.! !

!RPackage methodsFor: 'testing' stamp: 'tg 3/13/2010 13:47'!
includesExtensionSelector: aSelector ofClass: aClass
	"Return true if the receiver includes the method of selector aSelector. Only checks methods extending other packages"
	
	| sels |
	sels := self extensionSelectorsForClass: aClass.
	^ sels includes: aSelector.! !

!RPackage methodsFor: 'testing' stamp: 'StephaneDucasse 2/21/2010 18:54'!
includesExtensionSelector: aSelector ofClassName: aClassName
	"Return true if the receiver includes the method of selector aSelector. Only checks methods extending other packages"
	
	| sels |
	sels := classExtensionSelectors at: aClassName ifAbsent: [^ false].
	^ sels includes: aSelector.! !

!RPackage methodsFor: 'testing' stamp: 'StephaneDucasse 2/21/2010 18:54'!
includesExtensionSelector: aSelector ofMetaclassName: aClassName
	"Return true if the receiver includes the method of selector aSelector. Only checks methods extending other packages"
	
	| sels |
	sels := metaclassExtensionSelectors at: aClassName ifAbsent: [^ false].
	^ sels includes: aSelector.! !

!RPackage methodsFor: 'testing' stamp: 'tg 3/13/2010 16:42'!
includesSelector: aSelector ofClass: aClass
	"Return true if the receiver includes the method of selector aSelector. Checks methods defined locally as well as extending other packages"
	
	^ aClass isMeta
		ifTrue: [ self includesSelector: aSelector ofMetaclassName: aClass theNonMetaClass name  ]
		ifFalse: [ self includesSelector: aSelector ofClassName: aClass name  ]
	
! !

!RPackage methodsFor: 'testing' stamp: 'StephaneDucasse 2/21/2010 20:50'!
includesSelector: aSelector ofClassName: aClassName
	"Return true if the receiver includes the method of selector aSelector. Checks methods defined locally as well as extending other packages"
	
	^ (self includesDefinedSelector: aSelector ofClassName: aClassName)
		or: [(self includesExtensionSelector: aSelector ofClassName: aClassName)]! !

!RPackage methodsFor: 'testing' stamp: 'StephaneDucasse 3/11/2010 17:38'!
includesSelector: aSelector ofMetaclassName: aClassName
	"Return true if the receiver includes the method of selector aSelector. Checks methods defined locally as well as extending other packages"
	
	^ (self includesDefinedSelector: aSelector ofMetaclassName: aClassName)
		or: [(self includesExtensionSelector: aSelector ofMetaclassName: aClassName)]! !

!RPackage methodsFor: 'initialize-release' stamp: 'StephaneDucasse 3/11/2010 15:09'!
initialize
	super initialize.
	classes := Set new.
	metaclassExtensionSelectors := Dictionary new.
	classExtensionSelectors := Dictionary new.
	classDefinedSelectors := Dictionary new.
	metaclassDefinedSelectors := Dictionary new.! !

!RPackage methodsFor: 'slices' stamp: 'StephaneDucasse 3/11/2010 18:47'!
metaclassDefinedSlicesDo: aBlock
	"This method iterates over the defined class and their associated selectors. a slice is a class * list of selectors. aBlock will be applied to all the extensions slices of the receiver. aBlok first argument is the class and the second argument a list of method selectors"
	
	metaclassDefinedSelectors keysAndValuesDo: aBlock! !

!RPackage methodsFor: 'slices' stamp: 'StephaneDucasse 4/20/2010 21:42'!
metaclassExtensionSlicesDo: aBlock
	"This method iterates over the metaclass extensions and their associated selectors. A slice is a class * list of selectors. aBlock will be applied to all the extensions slices of the receiver. aBlok first argument is the class and second argument a list of selectors"
	
	metaclassExtensionSelectors keysAndValuesDo: aBlock
	
! !

!RPackage methodsFor: 'all accessing' stamp: 'sd 4/21/2010 16:36'!
methods

	| mths |
	mths := OrderedCollection new.
	metaclassExtensionSelectors keysAndValuesDo: [:key :val | val do: [:sel | mths add: ((Smalltalk at: key) class >> sel)]].
	classExtensionSelectors keysAndValuesDo: [:key :val |val do: [:sel | mths add: ((Smalltalk at: key) >> sel)]].
	metaclassDefinedSelectors keysAndValuesDo: [:key :val | val do: [:sel | mths add: ((Smalltalk at: key) class >> sel)]].
	classDefinedSelectors keysAndValuesDo: [:key :val | val do: [:sel | mths add: ((Smalltalk at: key) >> sel)]].
	^ mths! !

!RPackage methodsFor: 'accessing' stamp: 'StephaneDucasse 4/21/2010 14:05'!
methodsForClass: aClass
	"returns all the compiled for a class in the receiver without distinction of extensions or not"
	
	^ (self includesClass: aClass)
		ifFalse: [self extensionMethodsForClass: aClass]
		ifTrue: [self definedMethodsForClass: aClass]
	! !

!RPackage methodsFor: 'accessing' stamp: 'sd 4/25/2010 15:52'!
name

	^ name! !

!RPackage methodsFor: 'accessing' stamp: 'sd 4/25/2010 15:52'!
name: aSymbol

	name := aSymbol asSymbol! !

!RPackage methodsFor: 'printing' stamp: 'sd 4/25/2010 15:51'!
printOn: aStream
	super printOn: aStream.
	aStream nextPut: Character space.
	aStream nextPutAll: self name.! !

!RPackage methodsFor: 'private-register' stamp: 'tg 6/13/2010 17:29'!
registerClass: aClass
	self class organizer
		registerPackage: self forClass: aClass
	! !

!RPackage methodsFor: 'private-register' stamp: 'tg 6/13/2010 17:29'!
registerClassName: aClassNameSymbol
	self class organizer
		registerPackage: self forClassName: aClassNameSymbol
	! !

!RPackage methodsFor: 'class addition' stamp: 'sd 4/25/2010 15:47'!
removeClassDefinition: aClass
	"Remove the class definition from the package but not its methods"
	
	| removed aClassName|
	aClassName := aClass theNonMetaClass name.
	removed := classes remove: aClassName ifAbsent: [nil].
	removed ifNotNil: [
			self definedMethodsShouldBecomeExtensionWhenRemovingClass: aClassName.
			self unregisterClass: aClass.] 
	! !

!RPackage methodsFor: 'class addition' stamp: 'sd 4/25/2010 15:46'!
removeClassDefinitionName: aClassName
	"remove the class definition from the package but not its method"
	
	| removed aClassNameSymbol |
	('* class' match: aClassName) ifTrue: [^ self error: 'no metaclass name'].
	aClassNameSymbol := aClassName asSymbol.
	removed := classes remove: aClassNameSymbol ifAbsent: [nil].
	removed ifNotNil: [
				self definedMethodsShouldBecomeExtensionWhenRemovingClass: aClassName.
				self unregisterClassName: aClassNameSymbol.] 
	! !

!RPackage methodsFor: 'compiled method handling' stamp: 'tg 6/13/2010 17:44'!
removeMethod: aCompiledMethod 
	"Remove the method to the receiver as a defined method."

	| mr mthClass |
	mthClass := aCompiledMethod methodClass. 
	(self includesClass: mthClass)
		ifFalse: [mthClass isMeta
					ifTrue: [(metaclassExtensionSelectors 
								at: mthClass theNonMetaClass name  
								ifAbsentPut: [Set new]) remove: aCompiledMethod selector.]
					ifFalse: [(classExtensionSelectors 
								at: mthClass name 
								ifAbsentPut: [Set new]) remove: aCompiledMethod selector].
				"we remove a method extension so the receiver is an extending package of the class"
				RPackage organizer unregisterExtendingPackage: self forClass: mthClass]
		ifTrue: [mthClass isMeta
					ifTrue: [(metaclassDefinedSelectors 
								at: mthClass theNonMetaClass name  
								ifAbsentPut: [Set new]) remove: aCompiledMethod selector]
					ifFalse: [(classDefinedSelectors 
								at: mthClass name 
								ifAbsentPut: [Set new]) remove: aCompiledMethod selector]].	
	^ aCompiledMethod.! !

!RPackage methodsFor: 'selector handling' stamp: 'StephaneDucasse 4/20/2010 22:57'!
removeSelector: aSelector ofClassName: aClassName
	"Remove the method in the package but does not propagate to the class itself.  Note that this method does not remove the method from the class, it just records in the package that the method is not in the package anymore."
	
	( self includesClassNamed: aClassName)
		ifFalse: [(classExtensionSelectors  at: aClassName ifAbsent: [ ^ Set new]) remove: aSelector.]
		ifTrue: [(classDefinedSelectors  at: aClassName ifAbsent: [ ^ Set new]) remove: aSelector.].
	
	! !

!RPackage methodsFor: 'selector handling' stamp: 'StephaneDucasse 4/20/2010 22:57'!
removeSelector: aSelector ofMetaclassName: aClassName
	"Remove the method in the package. Note that this method does not remove the method from the class, it just records in the package that the method is not in the package anymore. aClassName is the sole instance class name and not its metaclass one: i.e. adding Point class>>new is done as removeSelector: #new ofMetaclassName: #Point"
	
	( self includesClassNamed: aClassName)
		ifFalse: [(metaclassExtensionSelectors  at: aClassName ifAbsent: [ ^ Set new]) remove: aSelector.]
		ifTrue: [(metaclassDefinedSelectors  at: aClassName ifAbsent: [ ^ Set new]) remove: aSelector.].
	
	! !

!RPackage methodsFor: 'all accessing' stamp: 'sd 4/21/2010 16:12'!
selectors

	| sel |
	sel := Set new.
	metaclassExtensionSelectors valuesDo: [:each | sel addAll: each].
	classExtensionSelectors valuesDo: [:each | sel addAll: each].
	metaclassDefinedSelectors valuesDo: [:each | sel addAll: each].
	classDefinedSelectors valuesDo: [:each | sel addAll: each].
	^ sel
	! !

!RPackage methodsFor: 'accessing' stamp: 'StephaneDucasse 4/21/2010 13:56'!
selectorsForClass: aClass
	"returns all the selector for a class in the receiver without distinction of extensions or not"
	
	^ (self includesClass: aClass)
		ifFalse: [self extensionSelectorsForClass: aClass]
		ifTrue: [self definedSelectorsForClass: aClass]
	! !

!RPackage methodsFor: 'private-register' stamp: 'tg 6/13/2010 17:29'!
unregisterClass: aClass
	self class organizer
		unregisterPackage: self forClass: aClass.! !

!RPackage methodsFor: 'private-register' stamp: 'tg 6/13/2010 17:29'!
unregisterClassName: aClassNameSymbol
	self class organizer
		unregisterPackage: self forClassName: aClassNameSymbol.! !

Object subclass: #RPackageBenchmark
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'RPackage-Core'!

!RPackageBenchmark class methodsFor: 'as yet unclassified' stamp: 'tg 6/5/2010 18:40'!
initializeRepository

	[	RPackage packageOrganizerClass: RMockPackageOrganizer.
		RMockPackageOrganizer reset ] 
	ensure: [ RPackage packageOrganizerClass: RPackageOrganizer ]! !

!RPackageBenchmark class methodsFor: 'as yet unclassified' stamp: 'tg 6/5/2010 18:40'!
run
	" self run "
	| classes |
	[
		RPackage packageOrganizerClass: RMockPackageOrganizer.
		RMockPackageOrganizer reset.
		classes := Object withAllSubclasses.
		classes remove: BogusInfo.
		classes remove: UndefinedObject.
		classes remove: ClassTrait.
		classes remove: ClassTrait class.
		classes remove: GRPackage.
		classes := classes reject: [:each |  'MC*' match: each name].
		Time millisecondsToRun: [
			classes do: [:each | 
				each package.
				each extendingPackages. 
				each package selectorsForClass: each.
				each extendingPackages do: [ :package | 
					package selectorsForClass: each.]. 
				each methods do: [:method | 
					method packageFromOrganizer: RMockPackageOrganizer default
					]
			]
		].
	] ensure: [ RPackage packageOrganizerClass: RPackageOrganizer ]! !

Object subclass: #RPackageOrganizer
	instanceVariableNames: 'classPackageMapping packages classExtendingPackagesMapping'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'RPackage-Core'!

RPackageOrganizer class
	instanceVariableNames: 'default'!

!RPackageOrganizer commentStamp: 'StephaneDucasse 5/15/2010 20:08' prior: 0!
A RPackageOrganizer is responsible for providing all the package currently defined in the system.
In addition it provides a back pointer from the class to its package.

The classPackageMapping and the classExtendingPackageMapping should be moved in the future to the classes
themselves. 

self systemChangeNotifier notify: self ofAllSystemChangesUsing: #callthismethodOnSelf.!

RPackageOrganizer class
	instanceVariableNames: 'default'!

!RPackageOrganizer class methodsFor: 'singleton' stamp: 'StephaneDucasse 5/15/2010 19:33'!
default
	
	default isNil ifTrue: [
		default := self new.
		default initializeFromSystem.
		default ].
	^ default! !

!RPackageOrganizer class methodsFor: 'singleton' stamp: 'tg 6/5/2010 18:40'!
fillUp
	"self fillUp"
	| package |
	PackageOrganizer default packages do: [ :info |
			package := RPackage named: info name.
			info classes do: [ :each | package addClassDefinition: each ].
			info coreMethods do: [ :each |
				each isValid
					ifTrue: [ package addMethod: each compiledMethod ] ].
			info extensionMethods do: [ :each |
				each isValid
					ifTrue: [ package addMethod: each compiledMethod ] ] ]
		displayingProgress: 'Importing'! !

!RPackageOrganizer methodsFor: 'package access from class' stamp: 'sd 5/21/2010 14:56'!
extendingPackagesOf: aClass
	"Returns the packages extending the class aClass"
	^ classExtendingPackagesMapping at: aClass theNonMetaClass name ifAbsent: [#()]! !

!RPackageOrganizer methodsFor: 'testing' stamp: 'sd 5/21/2010 15:21'!
includesClass: aClass

	^ self includesClassNamed: aClass name! !

!RPackageOrganizer methodsFor: 'testing' stamp: 'sd 5/21/2010 15:18'!
includesClassNamed: aSymbol

	^ classPackageMapping keys includes: aSymbol ! !

!RPackageOrganizer methodsFor: 'testing' stamp: 'sd 5/21/2010 15:17'!
includesPackage: aPackage

	^ self includesPackageNamed: aPackage name! !

!RPackageOrganizer methodsFor: 'testing' stamp: 'sd 5/21/2010 15:16'!
includesPackageNamed: aSymbol

	^ packages keys includes:  aSymbol ! !

!RPackageOrganizer methodsFor: 'initialize' stamp: 'StephaneDucasse 5/15/2010 22:22'!
initialize
	super initialize.
	packages := IdentityDictionary new.
	classPackageMapping := IdentityDictionary new.
	classExtendingPackagesMapping := IdentityDictionary new.
	self register! !

!RPackageOrganizer methodsFor: 'initialize' stamp: 'sd 5/21/2010 13:20'!
initializeFromSystem
	| package |
	MCWorkingCopy allManagers
		do: [ :workingCopy |
			| info |
			info := workingCopy packageInfo.
			package := self registerPackageNamed: info packageName.
			info classes 
				do: [ :each | package addClassDefinition: each ].
			info coreMethods do: [ :each |
				each isValid
					ifTrue: [ package addMethod: each compiledMethod ] ].
			info extensionMethods do: [ :each |
				each isValid
					ifTrue: [ package addMethod: each compiledMethod ] ] ]
		displayingProgress: 'Importing from MC managers'! !

!RPackageOrganizer methodsFor: 'accessing' stamp: 'tg 3/13/2010 16:39'!
packageDefiningOrExtendingMethod: aCompiledMethod
	^ self packageDefiningOrExtendingSelector: aCompiledMethod selector inClass: aCompiledMethod methodClass
! !

!RPackageOrganizer methodsFor: 'accessing' stamp: 'tg 6/5/2010 11:38'!
packageDefiningOrExtendingSelector: aSelector inClass: aClass
	"this implementation is slower
		aClass packages detect: [:each | each includesSelector: aSelector ofClass: aClass ]"

	^ aClass packages detect: [:each | each includesSelector: aSelector ofClass: aClass ]
	
"	self packages detect: [:each | 
		each includesSelector: aSelector ofClass: aClass ].
"	
	! !

!RPackageOrganizer methodsFor: 'accessing' stamp: 'stephane.ducasse 7/16/2009 23:17'!
packageNames
	^ packages keys! !

!RPackageOrganizer methodsFor: 'package access from class' stamp: 'tg 3/13/2010 16:28'!
packageOf: aClass

	self flag: #shouldProbablyReturnUnunpackagedPackage.
	^ classPackageMapping at: aClass theNonMetaClass name ifAbsent: [nil]! !

!RPackageOrganizer methodsFor: 'accessing' stamp: 'stephane.ducasse 7/10/2009 11:27'!
packages
	
	^ packages values! !

!RPackageOrganizer methodsFor: 'initialize' stamp: 'tg 6/13/2010 15:56'!
register
	MCWorkingCopy addDependent: self.
	self flag: 'to add listening to system announcer'.
"	SystemAnnouncer categoryAnnouncer when: SystemCategoryAddedAnnouncement do: [:ann |
		
		].
	SystemAnnouncer categoryAnnouncer when: SystemCategoryRemovedAnnouncement do: [:ann |
		
		]"! !

!RPackageOrganizer methodsFor: 'registration' stamp: 'sd 5/21/2010 14:44'!
registerExtendingPackage: aPackage forClass: aClass 

	| cur |
	cur := classExtendingPackagesMapping at: aClass theNonMetaClass name ifAbsent: [nil].
	cur isNil
		ifTrue: [classExtendingPackagesMapping at: aClass theNonMetaClass name put: (Set with: aPackage)]
		ifFalse: [cur add: aPackage]! !

!RPackageOrganizer methodsFor: 'registration' stamp: 'abc 5/9/2010 22:40'!
registerPackage: aPackage
	"Transcript show: self class printString , ' registerPackage: ', aPackage packageName asString ; cr."
	^ packages at: aPackage name asSymbol put: aPackage.
	! !

!RPackageOrganizer methodsFor: 'registration' stamp: 'TestRunner 12/25/2009 22:13'!
registerPackage: aPackage forClass: aClass 
	(aPackage includesClass: aClass) 
		ifFalse: 
			[self error: aPackage name , ' does not includes the class ' , aClass name].
	^classPackageMapping at: aClass theNonMetaClass name put: aPackage! !

!RPackageOrganizer methodsFor: 'registration' stamp: 'TestRunner 12/3/2009 21:42'!
registerPackage: aPackage forClassName: aClassNameSymbol
	(aPackage includesClassNamed: aClassNameSymbol asSymbol) 
		ifFalse: 
			[self error: aPackage name , ' does not includes the class ' , aClassNameSymbol].
	^classPackageMapping at: aClassNameSymbol put: aPackage! !

!RPackageOrganizer methodsFor: 'accessing' stamp: 'tg 6/5/2010 18:40'!
registerPackageNamed: aString
	^ RPackage named: aString! !

!RPackageOrganizer methodsFor: 'private' stamp: 'sd 5/21/2010 14:51'!
stopNotification
	"(self instVarNamed: #default) stopNotification"
	
	"pay attention that we can break the system using this method"
	
	SystemChangeNotifier uniqueInstance noMoreNotificationsFor: self.
	MCWorkingCopy removeDependent: self.! !

!RPackageOrganizer methodsFor: 'registration' stamp: 'StephaneDucasse 12/25/2009 19:47'!
unregisterClass: aClass 
	
	^ classPackageMapping removeKey: aClass theNonMetaClass name asSymbol
	! !

!RPackageOrganizer methodsFor: 'registration' stamp: 'sd 5/19/2010 16:13'!
unregisterExtendingPackage: aPackage forClass: aClass
	
	| cur |
	cur := classExtendingPackagesMapping at: aClass theNonMetaClass name ifAbsent: [nil].
	cur isNil
		ifFalse: [cur remove: aPackage ifAbsent: [] "not happy with this one"]! !

!RPackageOrganizer methodsFor: 'registration' stamp: 'sd 4/25/2010 15:51'!
unregisterPackage: aPackage
	"Transcript show: self class printString , ' unregisterPackage: ', aPackage packageName asString ; cr."
	
	packages removeKey: aPackage name asSymbol ifAbsent: [].	
	! !

!RPackageOrganizer methodsFor: 'registration' stamp: 'sd 5/19/2010 15:46'!
unregisterPackage: aPackage forClass: aClass 

	^classPackageMapping removeKey: aClass theNonMetaClass name! !

!RPackageOrganizer methodsFor: 'registration' stamp: 'TestRunner 12/25/2009 22:03'!
unregisterPackage: aPackage forClassName: aClassName
	(aPackage includesClassNamed: aClassName asSymbol) 
		ifTrue: 
			[self error: aPackage name , ' still includes the class ' , aClassName asSymbol].
	^classPackageMapping removeKey: aClassName asSymbol! !

!RPackageOrganizer methodsFor: 'registration' stamp: 'StephaneDucasse 12/20/2009 11:22'!
unregisterPackageNamed: aSymbolName
	packages removeKey: aSymbolName asSymbol ifAbsent: [].	
	! !

!RPackageOrganizer methodsFor: 'update' stamp: 'sd 5/21/2010 14:49'!
update: anArray
	Transcript show: 'From MCWorking copy'; cr.
	super update: anArray.
	(anArray isArray and: [ anArray first = #registered ])
		ifTrue: [ self registerPackageNamed: anArray second name ].
	(anArray isArray and: [ anArray first = #unregistered ])
		ifTrue: [ self unregisterPackageNamed: anArray second name ]! !
RPackage initialize!
